#pragma check_stack(off)
#include <stdlib.h>
#ifdef __BORLANDC__
#include <mem.h>
#else
#include <memory.h>
#endif

#include "systypes.h"
#define  DECL_DEVICES
#include "diskdev.h"
#include "biosdisk.h"
#include "debug.h"

/*
 * This module implements the concept of devices. It reads the partition
 * tables and creates an global array for all devices; this list
 * is extern'ed in the header file. Read and writesector routines may be 
 * used for accessing an individual partition. By setting a blocksize, you
 * can also read and write blocks.
 */


int fs_type_list[]={1,4,5,6,7,0x0A,0x51,0x64,0x80,0x81,0x82,0x83,0x93,0x94,0};
char *fs_type_name[]={"DOS 12-bit FAT","DOS 16-bit <32M", "Extended",
					  "DOS 16-bit >=32", "OS/2 HPFS","OS/2 Boot Manag",
					  "Novell?","Novell", "Old MINIX","Linux/MINIX",
					  "Linux swap","Linux native", "Amoeba","Amoeba BBT",
					  "Unknown"};

static char *find_fstype(int type)
{
	int	i=0;
	for(i=0; fs_type_list[i] && fs_type_list[i]!=type ; i++)
		;
	return fs_type_name[i];
}

/*
 * Create devices for each logical partition in an extended partition.
 * The logical partitions form a linked list, with each entry being
 * a partition table with two entries.  The first entry is the real data 
 * partition (with a start relative to the partition table start). The 
 * second is a pointer to the next logical partition (with a start relative 
 * to the entire extended partition).
 */

static uint8 read_extended_partitions(uint8 disk,uint8 cur_par)
{
	partition_t 	*p;
	uint32 			first_sector, this_sector;
	char			*data;

	data=(char *)malloc(1024);

	first_sector = lDevices[disk][cur_par].start_sec;
	this_sector = first_sector;

	while (cur_par<MAX_LDEV && ldev_read(&lDevices[disk][cur_par],0, 1,data))
	{
		if (*(uint16 *) (data+510) != 0xAA55)
			break;				// could have been in the while condition,
								// but I rather not depend on ldev_read being
								// called first

			/*
			 * Process the first entry, which should be the real
			 * data partition.
			 */

		p = (partition_t *) (0x1BE + data);
		if (p->sys_ind == TYPE_EXT_PARTITION ||	!p->nr_sec)
			break;  /* shouldn't happen */

			// overwrite the info, it was only needed to read the tables
		lDevices[disk][cur_par].start_sec = this_sector + p->start_sec;
		lDevices[disk][cur_par].nr_sec=p->nr_sec;
		lDevices[disk][cur_par].type=p->sys_ind;
		lDevices[disk][cur_par].disk=disk;
		lDevices[disk][cur_par].partition=cur_par;

		dprintf(D_PARTITION,"    hd%c%c, %4lu MB, %s (%u)\n",
			(int) disk+'a',(int) cur_par+'0', p->nr_sec >> 11,
			find_fstype(p->sys_ind),(unsigned) p->sys_ind);

		cur_par++;
		p++;

			/*
			 * Process the second entry, which should be a link
			 * to the next logical partition.  Create a fake partition
			 * for it just long enough to get the next partition
			 * table. The fake partition will be reused for the real
			 * data partition.
			 */

		//p = (partition_t *) (0x1BE + data);
		if (p->sys_ind != TYPE_EXT_PARTITION ||	!p->nr_sec)
			break;  /* no more logicals in this partition */

		lDevices[disk][cur_par].start_sec = this_sector + p->start_sec;
		//lDevices[disk][cur_par].nr_sec=p->nr_sec;
		lDevices[disk][cur_par].type=p->sys_ind;
		lDevices[disk][cur_par].disk=disk;
		lDevices[disk][cur_par].partition=cur_par;

		this_sector = first_sector + p->start_sec;
	}
	free(data);
	return cur_par;
}


/*
 * There are at most 4 primary partitions, they reside in the first
 * sector of a disk. If a primary partition's type is 6, it's an extended
 * partition. We print each partion with flag D_PARTITION, in this way
 * all partitions are printed when the debuglevel includes this bit.
 */

static void read_all_partitions(uint8 disk)
{
	int 				i;
	uint8				extended=5;
	partition_t    		*p;
	char				*data;

	data=(char *)malloc(1024);


	memset(lDevices[disk],0,sizeof(ldev_t)*MAX_LDEV);

	if (!pdev_read((uint8) disk,0, 1,data))
		return;


		// check for signature
	if (*(uint16 *) (data+510) == 0xAA55)
	{
		p = (partition_t *) (0x1BE + data);
		for (i=1 ; i<=4 ; i++,p++)
		{
			if (!p->nr_sec)
				continue;

			dprintf(D_PARTITION,"hd%c%c, %4lu MB, %s (%u)\n",
				(int) disk+'a',(int) i+'0', p->nr_sec >> 11,
				find_fstype(p->sys_ind),(unsigned) p->sys_ind);

			lDevices[disk][i].start_sec=p->start_sec;
			lDevices[disk][i].nr_sec=p->nr_sec;
			lDevices[disk][i].type=p->sys_ind;
			lDevices[disk][i].disk=disk;
			lDevices[disk][i].partition=i;

			if (p->sys_ind == TYPE_EXT_PARTITION)
			{
				/*
		 		 * Now handle the extended partitions. Make a new logical 
				 * sothat we can read the first partition table on it.
		 		 */
					
				lDevices[disk][extended].start_sec=p->start_sec;
				lDevices[disk][extended].disk=disk;
				lDevices[disk][extended].partition=extended;

				extended=read_extended_partitions(disk,extended);
			}
		}
	}
	free(data);
}

void pdev_readpartitions(void)
{
	uint8 	   		i;
	pdev_t			*p;

	for (i=0 ; i<MAX_HD ; i++)
	{
		p=&pDevices[i];
		bios_driveparam(i,&p->maxcyl,&p->maxsec,&p->maxhead);
		if (p->maxsec)
			read_all_partitions(i);
	}
}

int ldev_read(ldev_t *ldev, uint32 start_sec, uint8 nr_sec, void far *data)
{

	return pdev_read(ldev->disk,ldev->start_sec+start_sec,nr_sec,data);
}

int ldev_write(ldev_t *ldev,uint32 start_sec, uint8 nr_sec, void far * data)
{
	return pdev_write(ldev->disk,ldev->start_sec+start_sec,nr_sec,data);
}


int pdev_read(uint8 disk, uint32 start_sec, uint8 nr_sec, void far *data)
{
	uint32		sec,track,head,cyl;
	uint16		rval;

	sec   = start_sec % pDevices[disk].maxsec +1;
	track = start_sec / pDevices[disk].maxsec;
	head  = track % pDevices[disk].maxhead;
	cyl   = track / pDevices[disk].maxhead;

	rval=bios_rsect(disk, (uint8) nr_sec, (uint16) cyl, (uint8) sec, (uint8) head, data);
	return rval;

}

int pdev_write(uint8 disk, uint32 start_sec, uint8 nr_sec, void far * data)
{
	uint32		sec,track,head,cyl;

	sec   = start_sec % pDevices[disk].maxsec + 1;
	track = start_sec / pDevices[disk].maxsec;
	head  = track % pDevices[disk].maxhead;
	cyl   = track / pDevices[disk].maxhead;

	return bios_wsect(disk, (uint8) nr_sec, (uint16) cyl, (uint8) sec, (uint8) head, data);
}

int ldev_readblock(ldev_t *ldev, uint32 block, void far *data)
{
	dprintf(D_DISKDEV,"readblock, disk:%u, partition:%u, block:%lu\n",
		(unsigned) ldev->disk, (unsigned) ldev->partition, block);
	return ldev_read(ldev,block*ldev->sec_perblock,ldev->sec_perblock,data);
}

int ldev_writeblock(ldev_t *ldev, uint32 block, void far *data)
{
	dprintf(D_DISKDEV,"writeblock, disk:%u, partition:%u, block:%lu\n",
		(unsigned) ldev->disk, (unsigned) ldev->partition, block);
	return ldev_write(ldev,block*ldev->sec_perblock,ldev->sec_perblock,data);
}

int ldev_setblocksize(ldev_t *ldev, uint32 bsize)
{
	unsigned	sec_perblock = (unsigned) (bsize / (uint32) BYTES_PERSECTOR);

	if ( (uint32) sec_perblock * (uint32) BYTES_PERSECTOR != bsize)
	{
		dprintf(D_ALWAYS,"Blocksize %lu not supported on hd%c%c\n",bsize,(char) ldev->disk+'c',(char) ldev->partition);
		return 0;
	}
	ldev->sec_perblock=sec_perblock;
	dprintf(D_DISKDEV,"setblocksize, disk:%u, partition:%u, size:%lu\n",
		(unsigned) ldev->disk, (unsigned) ldev->partition, bsize);
	return 1;
}
