#pragma check_stack(off)
#include <stdlib.h>
#include <dos.h>

#include "debug.h"
#include "ems.h"
#include "emscache.h"
#include "tsr.h"

extern emshandle_t			emsCachePage;


/*
 * This module implements a cache buffer mechansism used
 * for inodes and diskblocks.
 * It is a two-level cache to prevent flushing when reading
 * "uninterresting" items. An item is lifted from level 1 to 2
 * after a cache-hit.
 *
 * Because data (ie inodes/diskblocks) resides in ems there is
 * a small problem when mapping different inodes/diskblocks at the
 * same time. That's why we allocate an extra emsPages in which we copy
 * the current mapped items; the emsPage is mapped in page 3 of the
 * emsPageFrame.
 * With cache_lookup we get the entry_cache of an element. If it resides
 * in the cache, 'data' points to the object, if not, it points (ie 'data')
 * to the entry in which we may store the new entry.
 * The mapping of cache objects is handled by a stack-mechanism,
 * 'cache_lookup' acts like "push", 'cache_release' acts like "pop".
 *
 * The stack is maintained by mapSP
 *
 * For efficiency we only handle objects witch a 2^n size!
 */



/* specify the number of items we idealy get on a queue */

#define HASH_QUEUE_SIZE 4
#define LEVEL1 '\1'
#define LEVEL2 '\2'

/* simple hashfun */
#define HASH_FUN(nr,sz)	((unsigned) ( nr % (unsigned long) sz))



/*
 *  remove entry, special care if were are the lru_head
 */

static void lru_remove(buffer_cache_t *bcache, cache_entry_t *entry)
{
	cache_entry_t **lru= &(bcache->lru_head[entry->level-1]);

	if (entry==(*lru))
		(*lru)=(*lru)->lru_next;
	entry->lru_prev->lru_next=entry->lru_next;
	entry->lru_next->lru_prev=entry->lru_prev;
}


/*
 * insert entry e1 before e2
 */
static void lru_putbefore(cache_entry_t *e1,cache_entry_t *e2)
{
	e1->lru_next=e2;
	e1->lru_prev=e2->lru_prev;
	e2->lru_prev->lru_next=e1;
	e2->lru_prev=e1;
}



static void hash_remove(cache_entry_t **hashqueue, cache_entry_t* entry)
{
	if (entry->hash_next==entry)
			/* removal of last entry */
		hashqueue[entry->hashindex]=0;
	else
	{
		entry->hash_next->hash_prev=entry->hash_prev;
		entry->hash_prev->hash_next=entry->hash_next;
			/* ensure that that queue points to something */
		hashqueue[entry->hashindex]=entry->hash_next;
	}
}

static void hash_add(cache_entry_t **hashqueue, cache_entry_t* entry)
{
	cache_entry_t	*hash_first=hashqueue[entry->hashindex];

	if (hash_first)
	{
		entry->hash_next=hash_first;
		entry->hash_prev=hash_first->hash_prev;
		hash_first->hash_prev->hash_next=entry;
		hash_first->hash_prev=entry;
	}
	else
	{
		entry->hash_next=entry;
		entry->hash_prev=entry;
	}

	hashqueue[entry->hashindex]=entry;
}



/*
 * We have an entry, put it last in the list (ie the most recent one)
 */

static void cache_update(buffer_cache_t *bcache,cache_entry_t *entry)
{

	lru_remove(bcache,entry);
	lru_putbefore(entry,bcache->lru_head[entry->level-1]);
}






/*
 * We replace 'entry' from level 1 with the lastest used entry from level2
 * because we made 'entry' the most recent in level 1, the level2_entry
 * will now be the most recent entry in level 1 (seems fair enough)
 *
 * Finally we mark 'entry' in level 2 as the most recent one
 */

static void cache_liftlevel(buffer_cache_t *bcache,cache_entry_t * entry)
{
	cache_entry_t 	*lru_head2=bcache->lru_head[1];
/*
 * first remove them
 */

	lru_remove(bcache,lru_head2);
	lru_remove(bcache,entry);

/*
 * switch level
 */

	lru_head2->level=LEVEL1;
	entry->level=LEVEL2;

/*
 * put them in the right position, that is most recently used
 */
	lru_putbefore(lru_head2,bcache->lru_head[0]);
	lru_putbefore(entry,bcache->lru_head[1]);
}






buffer_cache_t* mk_cache(unsigned emsPages,unsigned size_item)
{
	buffer_cache_t		*bcache;
	cache_entry_t		*entry;
	unsigned			i,halfway;

	if (! (bcache=(buffer_cache_t *)calloc(1,sizeof(buffer_cache_t))))
	{
		dprintf(D_ALWAYS,"Fatal could not alloc buffer_cache\n");
		goto error_init_cache_1;
	}
	Alloc_Size+=sizeof(buffer_cache_t);
	dprintf(D_SYSTEM,"Alloced: %u\n",sizeof(buffer_cache_t));


	bcache->nr_items=(PAGE_SIZE/size_item)*emsPages;
	bcache->size_item=size_item;
	if ( !(bcache->emsHandle=emsAllocPages(emsPages)))
	{
		dprintf(D_ALWAYS,"Could not alloc %u emspages for cache entries\n",emsPages);
		goto error_init_cache_2;
	}

	dprintf(D_CACHE,"Making cache with emshandle %u\n",bcache->emsHandle);
	if ( !(bcache->cache_list=(cache_entry_t*) calloc(bcache->nr_items,sizeof(cache_entry_t))))
	{
		dprintf(D_ALWAYS,"Could not alloc cache entries\n");
		goto error_init_cache_3;
	}
	Alloc_Size+=bcache->nr_items * sizeof(cache_entry_t);
	dprintf(D_SYSTEM,"Alloced cache entries: %u\n", bcache->nr_items * sizeof (cache_entry_t));
	bcache->lru_head[0]=bcache->cache_list;
	halfway=bcache->nr_items/2;
	bcache->lru_head[1]=bcache->lru_head[0]+halfway;

/*
 * Init level 1 cache
 */
	i=0;
	entry=bcache->lru_head[0];
	entry->lru_prev=bcache->lru_head[1]-1;
	for (; i<halfway ; i++,entry++)
	{
		entry->hash_next=entry->hash_prev=entry;
		entry->lru_next=entry+1;
		entry->level=LEVEL1;
		entry->emsIndex=i;
		(entry+1)->lru_prev=entry;
	}
	(entry-1)->lru_next=bcache->lru_head[0];

/*
 * Init level 2 cache
 */

	entry=bcache->lru_head[1];
	entry->lru_prev=bcache->lru_head[1]+halfway-1;
	for( ; i<bcache->nr_items ; i++,entry++)
	{
		entry->hash_next=entry->hash_prev=entry;
		entry->lru_next=entry+1;
		entry->level=LEVEL2;
		entry->emsIndex=i;
		(entry+1)->lru_prev=entry;
	}
	(entry-1)->lru_next=bcache->lru_head[1];

/*
 * alloc hashqueue, initialised with 0 pointers
 */

	bcache->hash_size=bcache->nr_items/HASH_QUEUE_SIZE;
	if ( !(bcache->hash_queue=(cache_entry_t **)calloc(bcache->hash_size,sizeof(cache_entry_t*))))
	{
		dprintf(D_ALWAYS,"Could not alloc hash_queue\n");
		goto error_init_cache_4;
	}
	Alloc_Size+=bcache->hash_size * sizeof(cache_entry_t *);
	dprintf(D_SYSTEM,"Alloced: %u\n", bcache->hash_size  * sizeof(cache_entry_t *));
	return bcache;

error_init_cache_4:
	free(bcache->cache_list);

error_init_cache_3:
	emsFreePages(bcache->emsHandle);

error_init_cache_2:
	free(bcache);

error_init_cache_1:
	return (buffer_cache_t*) 0;
}


/*
 * Delete a cache
 */


void dl_cache(buffer_cache_t *bcache)
{
	dprintf(D_CACHE,"Deleting cache with emshandle %u\n",bcache->emsHandle);
	dprintf(D_CACHE,"lookups: %lu, hits:%lu\n",bcache->lookups,bcache->hits);

	emsFreePages(bcache->emsHandle);
#ifndef RUNNING_AS_TSR
	free(bcache->cache_list);
	free(bcache->hash_queue);
	free(bcache);
#endif
}

/*
 * Lookup item.. returns 0 when not found
 */

cache_entry_t* cache_lookup(buffer_cache_t *bcache, unsigned l_unit, unsigned long item_no)
{

	cache_entry_t		*entry,*hash_first;

	bcache->lookups++;

	hash_first=bcache->hash_queue[ HASH_FUN(item_no,bcache->hash_size) ];
	if (!hash_first)
		return (cache_entry_t *)0;

	entry=hash_first;
	do
	{
		if (item_no == entry->item_no && l_unit==(unsigned) entry->l_unit)
		{
			bcache->hits++;
			if (entry->level==LEVEL1)
				cache_liftlevel(bcache,entry);
			else
				cache_update(bcache,entry);
			return entry;
		}
		else
			entry=entry->hash_next;

	} while (entry!=hash_first);

	return (cache_entry_t *) 0;
}




/*
 * Add an item to the cache, it is made the most recent in level 1
 */

cache_entry_t* cache_add(buffer_cache_t *bcache, unsigned l_unit, unsigned long item_no, void far *data)
{
	unsigned 			newhashindex;
	unsigned 			emsPage;
	unsigned long 		offset;
	cache_entry_t*		entry;

	entry=bcache->lru_head[0];
	entry->item_no=item_no;
	entry->l_unit=(unsigned char) l_unit;

	offset=(unsigned long) bcache->size_item * (unsigned long) entry->emsIndex;
	emsPage=(unsigned) (offset / (unsigned long) PAGE_SIZE);
	offset=offset % (unsigned long) PAGE_SIZE;
//	emsConv2Ems(data,bcache->emsHandle,emsPage,(unsigned) offset,bcache->size_item);
	emsEms2Ems(emsCachePage,0,FP_OFF(data)-0xC000,bcache->emsHandle,emsPage,(unsigned) offset,bcache->size_item);
	entry->data=0;

	newhashindex=HASH_FUN(item_no,bcache->hash_size);
	if (newhashindex!=entry->hashindex)
	{
		hash_remove(bcache->hash_queue,entry);
		entry->hashindex=newhashindex;
		hash_add(bcache->hash_queue,entry);
	}
	bcache->lru_head[0]=entry->lru_next;

	return entry;
}


void cache_walk(buffer_cache_t *bcache, int level)
{
	cache_entry_t		*first,*entry;

	entry=first=bcache->lru_head[level-1];
	do
	{
	} while ( (entry=entry->lru_next)!=first);
}
